﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  internal class M5RuleSetModel : RegressionModel, IM5Model {
    private const string NumRulesResultName = "Number of rules";
    private const string CoveredInstancesResultName = "Covered instances";

    #region Properties
    [Storable]
    internal List<M5RuleModel> Rules { get; private set; }
    #endregion

    #region HLConstructors & Cloning
    [StorableConstructor]
    protected M5RuleSetModel(bool deserializing) : base(deserializing) { }
    protected M5RuleSetModel(M5RuleSetModel original, Cloner cloner) : base(original, cloner) {
      if (original.Rules != null) Rules = original.Rules.Select(cloner.Clone).ToList();
    }
    protected M5RuleSetModel(string targetVariable) : base(targetVariable) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5RuleSetModel(this, cloner);
    }
    #endregion

    internal static M5RuleSetModel CreateRuleModel(string targetAttr, M5Parameters m5Params) {
      return m5Params.LeafModel.ProvidesConfidence ? new ConfidenceM5RuleSetModel(targetAttr) : new M5RuleSetModel(targetAttr);
    }

    #region RegressionModel
    public override IEnumerable<string> VariablesUsedForPrediction {
      get {
        var f = Rules.FirstOrDefault();
        return f != null ? (f.VariablesUsedForPrediction ?? new List<string>()) : new List<string>();
      }
    }
    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      if (Rules == null) throw new NotSupportedException("The model has not been built yet");
      return rows.Select(row => GetEstimatedValue(dataset, row));
    }
    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, problemData);
    }
    #endregion

    #region IM5Model
    public void Build(IReadOnlyList<int> trainingRows, IReadOnlyList<int> pruningRows, M5Parameters m5Params, CancellationToken cancellationToken) {
      Rules = new List<M5RuleModel>();
      var tempTraining = trainingRows;
      var tempPruning = pruningRows;
      do {
        var tempRule = M5RuleModel.CreateRuleModel(m5Params.TargetVariable, m5Params);
        cancellationToken.ThrowIfCancellationRequested();

        if (!m5Params.Results.ContainsKey(NumRulesResultName)) m5Params.Results.Add(new Result(NumRulesResultName, new IntValue(0)));
        if (!m5Params.Results.ContainsKey(CoveredInstancesResultName)) m5Params.Results.Add(new Result(CoveredInstancesResultName, new IntValue(0)));

        var t1 = tempTraining.Count;
        tempRule.Build(tempTraining, tempPruning, m5Params, cancellationToken);
        tempTraining = tempTraining.Where(i => !tempRule.Covers(m5Params.Data, i)).ToArray();
        tempPruning = tempPruning.Where(i => !tempRule.Covers(m5Params.Data, i)).ToArray();
        Rules.Add(tempRule);
        ((IntValue)m5Params.Results[NumRulesResultName].Value).Value++;
        ((IntValue)m5Params.Results[CoveredInstancesResultName].Value).Value += t1 - tempTraining.Count;
      }
      while (tempTraining.Count > 0);
    }

    public void Update(IReadOnlyList<int> rows, M5Parameters m5Parameters, CancellationToken cancellationToken) {
      foreach (var rule in Rules) rule.Update(rows, m5Parameters, cancellationToken);
    }
    #endregion

    #region Helpers
    private double GetEstimatedValue(IDataset dataset, int row) {
      foreach (var rule in Rules) {
        if (rule.Covers(dataset, row))
          return rule.GetEstimatedValues(dataset, row.ToEnumerable()).Single();
      }
      throw new ArgumentException("Instance is not covered by any rule");
    }
    #endregion

    [StorableClass]
    private class ConfidenceM5RuleSetModel : M5RuleSetModel, IConfidenceRegressionModel {
      #region HLConstructors & Cloning
      [StorableConstructor]
      protected ConfidenceM5RuleSetModel(bool deserializing) : base(deserializing) { }
      private ConfidenceM5RuleSetModel(ConfidenceM5RuleSetModel original, Cloner cloner) : base(original, cloner) { }
      public ConfidenceM5RuleSetModel(string targetVariable) : base(targetVariable) { }
      public override IDeepCloneable Clone(Cloner cloner) {
        return new ConfidenceM5RuleSetModel(this, cloner);
      }
      #endregion

      #region IConfidenceRegressionModel
      public IEnumerable<double> GetEstimatedVariances(IDataset dataset, IEnumerable<int> rows) {
        if (Rules == null) throw new NotSupportedException("The model has not been built yet");
        return rows.Select(row => GetEstimatedVariance(dataset, row));
      }
      public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
        return new ConfidenceRegressionSolution(this, problemData);
      }
      private double GetEstimatedVariance(IDataset dataset, int row) {
        foreach (var rule in Rules) {
          if (rule.Covers(dataset, row)) return ((IConfidenceRegressionModel)rule).GetEstimatedVariances(dataset, row.ToEnumerable()).Single();
        }
        throw new ArgumentException("Instance is not covered by any rule");
      }
      #endregion
    }
  }
}